"""Command implementation."""
from __future__ import annotations

from copy import deepcopy
from typing import Any, Callable, Optional

TypeProcessResultCb = Optional[Callable[[Any], Optional[str]]]


class Command:
    """The object for coap commands."""

    def __init__(
        self,
        method: str,
        path: list[str],
        data: dict[str, Any] | None = None,
        *,
        parse_json: bool = True,
        observe: bool = False,
        observe_duration: int = 0,
        process_result: TypeProcessResultCb = None,
        err_callback: Callable[[Exception], None] | None = None,
    ) -> None:
        """Create object of class."""
        self._method = method
        self._path = path
        self._data = data
        self._parse_json = parse_json
        self._process_result = process_result
        self._err_callback = err_callback
        self._observe = observe
        self._observe_duration = observe_duration
        self._raw_result: str | None = None
        self._result: str | None = None

    @property
    def method(self) -> str:
        """Return method."""
        return self._method

    @property
    def path(self) -> list[str]:
        """Return path."""
        return self._path

    @property
    def data(self) -> dict[str, Any] | None:
        """Return data."""
        return self._data

    @property
    def parse_json(self) -> bool:
        """Json parsing result."""
        return self._parse_json

    @property
    def process_result(self) -> TypeProcessResultCb:
        """Definition of callback to process result."""
        return self._process_result

    @property
    def err_callback(self) -> Callable[[Exception], None] | None:
        """Will be fired when an observe request fails."""
        return self._err_callback

    @property
    def observe(self) -> bool:
        """Observe function."""
        return self._observe

    @property
    def observe_duration(self) -> int:
        """Return duration period of observations."""
        return self._observe_duration

    @property
    def raw_result(self) -> str | None:
        """Return raw result."""
        return self._raw_result

    @property
    def result(self) -> str | None:
        """Return result."""
        return self._result

    @result.setter
    def result(self, value: str) -> None:
        """Return command result."""
        if self._process_result:
            self._result = self._process_result(value)

        self._raw_result = value

    @property
    def path_str(self) -> str:
        """Return coap path."""
        return "/".join(str(v) for v in self._path)

    def url(self, host: str) -> str:
        """Generate url for coap client."""
        return f"coaps://{host}:5684/{self.path_str}"

    def _merge(
        self, a_dict: dict[str, Any] | None, b_dict: dict[str, Any] | None
    ) -> dict[str, Any] | None:
        """Merge a into b."""
        if a_dict is None or b_dict is None:
            return None
        for key, value in a_dict.items():
            if isinstance(value, dict):
                item = b_dict.setdefault(key, {})
                self._merge(value, item)
            elif isinstance(value, list):
                item = b_dict.setdefault(key, [{}])
                if len(value) == 1 and isinstance(value[0], dict):
                    self._merge(value[0], item[0])
                else:
                    b_dict[key] = value
            else:
                b_dict[key] = value
        return b_dict

    def combine_data(self, command2: Command | None) -> None:
        """Combine data for this command with another."""
        if command2 is None:
            return
        self._data = self._merge(
            command2._data, self._data  # pylint: disable=protected-access
        )

    def __add__(self, other: Command | None) -> Command:
        """Add Command to this Command."""
        if other is None:
            return deepcopy(self)
        if isinstance(other, self.__class__):
            new_obj = deepcopy(self)
            new_obj.combine_data(other)
            return new_obj
        msg = f"unsupported operand type(s) for '{self.__class__}' and '{type(other)}'"
        raise (TypeError(msg))

    def __repr__(self) -> str:
        """Return the representation."""
        if self.data is None:
            template = "<Command {} {}{}>"
        else:
            template = "<Command {} {}: {}>"
        return template.format(self.method, self.path, self.data or "")



"""COAP implementation."""
from __future__ import annotations

from functools import wraps
import json
import logging
import subprocess
from time import time
from typing import TYPE_CHECKING

from ..command import Command
from ..error import ClientError, RequestError, RequestTimeout, ServerError
from ..gateway import Gateway

_LOGGER = logging.getLogger(__name__)

CLIENT_ERROR_PREFIX = "4."
SERVER_ERROR_PREFIX = "5."


class APIFactory:
    """APIFactory."""

    def __init__(self, host, psk_id="pytradfri", psk=None, timeout=10):
        """Create object of class."""
        self._host = host
        self._psk_id = psk_id
        self._psk = psk
        self._timeout = timeout  # seconds

    @property
    def psk(self):
        """Return psk."""
        return self._psk

    @psk.setter
    def psk(self, value):
        """Set psk."""
        self._psk = value

    def _base_command(self, method: str) -> list[str]:
        """Return base command."""
        return [
            "coap-client",
            "-u",
            self._psk_id,
            "-k",
            self._psk,
            "-v",
            "0",
            "-m",
            method,
        ]

    def _execute(self, api_command, *, timeout=None):
        """Execute the command."""

        if api_command.observe:
            self._observe(api_command)
            return None

        method = api_command.method
        path = api_command.path
        data = api_command.data
        parse_json = api_command.parse_json
        url = api_command.url(self._host)

        proc_timeout = self._timeout
        if timeout is not None:
            proc_timeout = timeout

        command = self._base_command(method)

        kwargs = {
            "stderr": subprocess.DEVNULL,
            "timeout": proc_timeout,
            "universal_newlines": True,
        }

        if data is not None:
            kwargs["input"] = json.dumps(data)
            command.append("-f")
            command.append("-")
            _LOGGER.debug("Executing %s %s %s: %s", self._host, method, path, data)
        else:
            _LOGGER.debug("Executing %s %s %s", self._host, method, path)

        command.append(url)

        try:
            return_value = subprocess.check_output(command, **kwargs)
        except subprocess.TimeoutExpired:
            raise RequestTimeout() from None
        except subprocess.CalledProcessError as exc:
            msg = f"Error executing request: {exc}"
            raise RequestError(msg) from None

        api_command.result = _process_output(return_value, parse_json)
        return api_command.result

    def request(self, api_commands, *, timeout=None):
        """Make a request. Timeout is in seconds."""
        if not isinstance(api_commands, list):
            return self._execute(api_commands, timeout=timeout)

        command_results = []

        for api_command in api_commands:
            result = self._execute(api_command, timeout=timeout)
            command_results.append(result)

        return command_results

    def _observe(self, api_command: Command) -> None:
        """Observe an endpoint."""
        path = api_command.path
        duration = api_command.observe_duration
        if duration <= 0:
            raise ValueError("Observation duration has to be greater than 0.")
        url = api_command.url(self._host)
        err_callback = api_command.err_callback

        command = self._base_command("get") + [
            "-s",
            str(duration),
            "-B",
            str(duration),
            url,
        ]

        try:
            proc = subprocess.Popen(  # pylint: disable=consider-using-with
                command,
                stdout=subprocess.PIPE,
                stderr=subprocess.DEVNULL,
                universal_newlines=True,
            )
        except subprocess.CalledProcessError as exc:
            msg = f"Error executing request: {exc}"
            raise RequestError(msg) from None

        output = ""
        open_obj = 0
        start = time()

        def read_stdout() -> str:
            """Read from stdout."""
            if TYPE_CHECKING:
                assert proc.stdout
            return proc.stdout.read(1)

        for data in iter(read_stdout, ""):
            if data == "\n":
                _LOGGER.debug(
                    "Observing stopped for %s after %.1fs", path, time() - start
                )
                if err_callback:
                    err_callback(RequestError("Observing stopped."))
                break

            if data == "{":
                open_obj += 1
            elif data == "}":
                open_obj -= 1

            output += data

            if open_obj == 0:
                api_command.result = _process_output(output)
                output = ""

    def generate_psk(self, security_key):
        """Generate and set a psk from the security key."""
        if not self._psk:
            # Backup the real identity.
            existing_psk_id = self._psk_id

            # Set the default identity and security key for generation.
            self._psk_id = "Client_identity"
            self._psk = security_key

            # Ask the Gateway to generate the psk for the identity.
            self._psk = self.request(Gateway().generate_psk(existing_psk_id))

            # Restore the real identity.
            self._psk_id = existing_psk_id

        return self._psk


def _process_output(output, parse_json=True):
    """Process output."""
    output = output.strip()
    _LOGGER.debug("Received: %s", output)

    if not output:
        return None
    if "decrypt_verify" in output:
        raise RequestError(
            "Please compile coap-client without debug output. See "
            "instructions at "
            "https://github.com/ggravlingen/pytradfri#installation"
        )
    if output.startswith(CLIENT_ERROR_PREFIX):
        raise ClientError(output)
    if output.startswith(SERVER_ERROR_PREFIX):
        raise ServerError(output)
    if not parse_json:
        return output
    return json.loads(output)


def retry_timeout(api, retries=3):
    """Retry API call when a timeout occurs."""

    @wraps(api)
    def retry_api(*args, **kwargs):
        """Retrying API."""
        for i in range(1, retries + 1):
            try:
                return api(*args, **kwargs)
            except RequestTimeout:
                if i == retries:
                    raise
        return None

    return retry_api



"""Smart tasks set timers to turn on/off lights in various ways.

> Currently supporting wake up

SmartTask # return top level info
    TaskControl # Change top level values
    StartAction # Get top level info on start action
        StartActionItem # Get info on specific device in task
            StartActionItemController # change values for task
"""

import datetime
from datetime import datetime as dt

from .command import Command
from .const import (
    ATTR_DEVICE_STATE,
    ATTR_ID,
    ATTR_LIGHT_DIMMER,
    ATTR_REPEAT_DAYS,
    ATTR_SMART_TASK_LIGHTS_OFF,
    ATTR_SMART_TASK_NOT_AT_HOME,
    ATTR_SMART_TASK_TRIGGER_TIME_INTERVAL,
    ATTR_SMART_TASK_TRIGGER_TIME_START_HOUR,
    ATTR_SMART_TASK_TRIGGER_TIME_START_MIN,
    ATTR_SMART_TASK_TYPE,
    ATTR_SMART_TASK_WAKE_UP,
    ATTR_START_ACTION,
    ATTR_TRANSITION_TIME,
    ROOT_SMART_TASKS,
    ROOT_START_ACTION,
)
from .resource import ApiResource
from .util import BitChoices

WEEKDAYS = BitChoices(
    (
        ("mon", "Monday"),
        ("tue", "Tuesday"),
        ("wed", "Wednesday"),
        ("thu", "Thursday"),
        ("fri", "Friday"),
        ("sat", "Saturday"),
        ("sun", "Sunday"),
    )
)


class SmartTask(ApiResource):
    """Represent a smart task."""

    def __init__(self, gateway, raw):
        """Initialize the class."""
        super().__init__(raw)
        self._gateway = gateway

    @property
    def path(self):
        """Return gateway path."""
        return [ROOT_SMART_TASKS, self.id]

    @property
    def state(self):
        """Boolean representing the light state of the transition."""
        return self.raw.get(ATTR_DEVICE_STATE) == 1

    @property
    def task_type_id(self):
        """Return type of task."""
        return self.raw.get(ATTR_SMART_TASK_TYPE)

    @property
    def task_type_name(self):
        """Return the task type in plain text.

        (Own interpretation of names.)
        """
        if self.is_wake_up:
            return "Wake Up"
        if self.is_not_at_home:
            return "Not At Home"
        if self.is_lights_off:
            return "Lights Off"
        return None

    @property
    def is_wake_up(self):
        """Boolean representing if this is a wake up task."""
        return self.raw.get(ATTR_SMART_TASK_TYPE) == ATTR_SMART_TASK_WAKE_UP

    @property
    def is_not_at_home(self):
        """Boolean representing if this is a not home task."""
        return self.raw.get(ATTR_SMART_TASK_TYPE) == ATTR_SMART_TASK_NOT_AT_HOME

    @property
    def is_lights_off(self):
        """Boolean representing if this is a lights off task."""
        return self.raw.get(ATTR_SMART_TASK_TYPE) == ATTR_SMART_TASK_LIGHTS_OFF

    @property
    def repeat_days(self):
        """Return int (bit) for enabled weekdays."""
        return self.raw.get(ATTR_REPEAT_DAYS)

    @property
    def repeat_days_list(self):
        """Binary representation of weekdays the event takes place."""
        return WEEKDAYS.get_selected_values(self.raw.get(ATTR_REPEAT_DAYS))

    @property
    def task_start_parameters(self):
        """Return hour and minute that task starts."""
        return self.raw.get(ATTR_SMART_TASK_TRIGGER_TIME_INTERVAL)[0]

    @property
    def task_start_time(self):
        """Return the time the task starts.

        Time is set according to iso8601.
        """
        return datetime.time(
            self.task_start_parameters[ATTR_SMART_TASK_TRIGGER_TIME_START_HOUR],
            self.task_start_parameters[ATTR_SMART_TASK_TRIGGER_TIME_START_MIN],
        )

    @property
    def task_control(self):
        """Control a task."""
        return TaskControl(self, self.state, self.path, self._gateway)

    @property
    def start_action(self):
        """Return start action object."""
        return StartAction(self, self.path)

    def __repr__(self):
        """Return a readable name for smart task."""
        state = "on" if self.state else "off"
        return f"<Task {self.id} - {self.task_type_name} - {state}>"


class TaskControl:
    """Class to control the tasks."""

    def __init__(self, task, state, path, gateway):
        """Initialize TaskControl."""
        self._task = task
        self.state = state
        self.path = path
        self._gateway = gateway

    @property
    def tasks(self):
        """Return task objects of the task control."""
        return [
            StartActionItem(self._task, i, self.state, self.path, self.raw)
            for i in range(len(self.raw))
        ]

    def set_dimmer_start_time(self, hour, minute):
        """Set start time for task (hh:mm) in iso8601.

        NB: dimmer starts 30 mins before time in app
        """
        #  This is to calculate the difference between local time
        #  and the time in the gateway
        d_now = self._gateway.get_gateway_info().current_time
        d_utcnow = dt.utcnow()
        diff = d_now - d_utcnow
        newtime = dt(100, 1, 1, hour, minute, 00) - diff

        command = {
            ATTR_SMART_TASK_TRIGGER_TIME_INTERVAL: [
                {
                    ATTR_SMART_TASK_TRIGGER_TIME_START_HOUR: newtime.hour,
                    ATTR_SMART_TASK_TRIGGER_TIME_START_MIN: newtime.minute,
                }
            ]
        }
        return self._task.set_values(command)

    @property
    def raw(self):
        """Return raw data that it represents."""
        return self._task.raw[ATTR_START_ACTION]


class StartAction:
    """Class to control the start action-node."""

    def __init__(self, start_action, path):
        """Initialize StartAction class."""
        self.start_action = start_action
        self.path = path

    @property
    def state(self):
        """Return state of start action task."""
        return self.raw.get(ATTR_DEVICE_STATE)

    @property
    def devices(self):
        """Return state of start action task."""
        return [
            StartActionItem(self.start_action, i, self.state, self.path, self.raw)
            for i in range(len(self.raw[ROOT_START_ACTION]))
        ]

    @property
    def raw(self):
        """Return raw data that it represents."""
        return self.start_action.raw[ATTR_START_ACTION]


class StartActionItem:
    """Class to show settings for a task."""

    def __init__(self, task, index, state, path, raw):
        """Initialize TaskInfo."""
        self.task = task
        self.index = index
        self.state = state
        self.path = path
        self._raw = raw

    @property
    def devices_dict(self):
        """Return state of start action task."""
        json_list = {}
        index = 0
        for x_entry in self._raw[ROOT_START_ACTION]:
            if index != self.index:
                json_list.update(x_entry)
            index = index + 1
        return json_list

    @property
    def id(self):
        """Return ID (device id) of task."""
        return self.raw.get(ATTR_ID)

    @property
    def item_controller(self):
        """Control a task."""
        return StartActionItemController(
            self, self.raw, self.state, self.path, self.devices_dict
        )

    @property
    def transition_time(self):
        """Transition runs for this long from the time in task_start.

        Value is in seconds x 10. Default to 0 if transition is missing.
        """
        return self.raw.get(ATTR_TRANSITION_TIME, 0) / 60 / 10

    @property
    def dimmer(self):
        """Return dimmer level."""
        return self.raw.get(ATTR_LIGHT_DIMMER)

    @property
    def raw(self):
        """Return raw data that it represents."""
        return self._raw[ROOT_START_ACTION][self.index]

    def __repr__(self):
        """Return a readable name for this class."""
        return f"<StartActionItem (Device: {self.id} - Dimmer: {self.dimmer} - Time: {self.transition_time})>"


class StartActionItemController:
    """Class to edit settings for a task."""

    def __init__(self, item, raw, state, path, devices_dict):
        """Initialize TaskControl."""
        self._item = item
        self.raw = raw
        self.state = state
        self.path = path
        self.devices_dict = devices_dict

    def set_dimmer(self, dimmer):
        """Set final dimmer value for task."""
        command = {
            ATTR_START_ACTION: {
                ATTR_DEVICE_STATE: self.state,
                ROOT_START_ACTION: [
                    {
                        ATTR_ID: self.raw[ATTR_ID],
                        ATTR_LIGHT_DIMMER: dimmer,
                        ATTR_TRANSITION_TIME: self.raw.get(ATTR_TRANSITION_TIME, 0),
                    },
                    self.devices_dict,
                ],
            }
        }
        return self.set_values(command)

    def set_transition_time(self, transition_time):
        """Set time (mins) for light transition."""
        command = {
            ATTR_START_ACTION: {
                ATTR_DEVICE_STATE: self.state,
                ROOT_START_ACTION: [
                    {
                        ATTR_ID: self.raw[ATTR_ID],
                        ATTR_LIGHT_DIMMER: self.raw[ATTR_LIGHT_DIMMER],
                        ATTR_TRANSITION_TIME: transition_time * 10 * 60,
                    },
                    self.devices_dict,
                ],
            }
        }
        return self.set_values(command)

    def set_values(self, command):
        """
        Set values on task control.

        Returns a Command.
        """
        return Command("put", self._item.path, command)



"""Represent an air purifier."""
from __future__ import annotations

from typing import TYPE_CHECKING, TypedDict, cast

from ..const import (
    ATTR_AIR_PURIFIER_AIR_QUALITY,
    ATTR_AIR_PURIFIER_CONTROLS_LOCKED,
    ATTR_AIR_PURIFIER_FAN_SPEED,
    ATTR_AIR_PURIFIER_FILTER_LIFETIME_REMAINING,
    ATTR_AIR_PURIFIER_FILTER_LIFETIME_TOTAL,
    ATTR_AIR_PURIFIER_FILTER_RUNTIME,
    ATTR_AIR_PURIFIER_FILTER_STATUS,
    ATTR_AIR_PURIFIER_LEDS_OFF,
    ATTR_AIR_PURIFIER_MODE,
    ATTR_AIR_PURIFIER_MODE_AUTO,
    ATTR_AIR_PURIFIER_MOTOR_RUNTIME_TOTAL,
    ROOT_AIR_PURIFIER,
)
from ..resource import TypeRawList

if TYPE_CHECKING:
    # avoid cyclic import at runtime.
    from . import Device


TypeAirPurifier = TypedDict(
    # Alternative syntax required due to the need of using strings as keys:
    # https://www.python.org/dev/peps/pep-0589/#alternative-syntax
    "TypeAirPurifier",
    {
        "5900": int,  # Mode
        "5902": int,  # Filter runtume
        "5903": int,  # Filter status
        "5904": int,  # Filter lifetime total
        "5905": int,  # Manual controls locked
        "5906": int,  # Led light on/off
        "5907": int,  # Air quality level
        "5908": int,  # Fan speed
        "5909": int,  # Motor runtime total
        "5910": int,  # Filter lifetime remaining
        "9003": int,  # ID
    },
)


class AirPurifier:
    """Represent an air purifier."""

    def __init__(self, device: Device, index: int) -> None:
        """Create object of class."""
        self.device = device
        self.index = index

    @property
    def air_quality(self) -> int:
        """Get the current air quality measured by the air purifier.

        0..35: Good
        36..85: OK
        86..: Not good
        65535: If the fan is off or during measuring time after turning on
        """
        return self.raw[ATTR_AIR_PURIFIER_AIR_QUALITY]

    @property
    def controls_locked(self) -> bool:
        """Return True if physical controls on the air purifier are locked."""
        return self.raw[ATTR_AIR_PURIFIER_CONTROLS_LOCKED] == 1

    @property
    def fan_speed(self) -> int:
        """Get the current fan speed of the air purifier.

        0: Device is off
        2..50: Fan speed with a step size of 1.
        """
        return self.raw[ATTR_AIR_PURIFIER_FAN_SPEED]

    @property
    def filter_lifetime_remaining(self) -> int:
        """Return remaining lifetime of filter, expressed in minutes."""
        return self.raw[ATTR_AIR_PURIFIER_FILTER_LIFETIME_REMAINING]

    @property
    def filter_lifetime_total(self) -> int:
        """Return total lifetime of filter, expressed in minutes."""
        return self.raw[ATTR_AIR_PURIFIER_FILTER_LIFETIME_TOTAL]

    @property
    def filter_runtime(self) -> int:
        """Return filter runtime, expressed in minutes."""
        return self.raw[ATTR_AIR_PURIFIER_FILTER_RUNTIME]

    @property
    def filter_status(self) -> int:
        """Return filter status."""
        return self.raw[ATTR_AIR_PURIFIER_FILTER_STATUS]

    @property
    def is_auto_mode(self) -> bool:
        """
        Return auto mode on or off.

        Auto mode sets the fan speed automatically based on the air quality.
        """
        return self.raw[ATTR_AIR_PURIFIER_MODE] == ATTR_AIR_PURIFIER_MODE_AUTO

    @property
    def leds_off(self) -> bool:
        """Return True if led's on the air purifier are turned off."""
        return self.raw[ATTR_AIR_PURIFIER_LEDS_OFF] == 1

    @property
    def motor_runtime_total(self) -> int:
        """Return runtime of fan motor, expressed in minutes."""
        return self.raw[ATTR_AIR_PURIFIER_MOTOR_RUNTIME_TOTAL]

    @property
    def raw(self) -> TypeAirPurifier:
        """Return raw data that it represents."""
        return cast(
            TypeAirPurifier,
            cast(TypeRawList, self.device.raw)[ROOT_AIR_PURIFIER][self.index],
        )

    @property
    def state(self) -> bool:
        """Return device state, ie on or off."""
        return self.raw[ATTR_AIR_PURIFIER_MODE] > 0



"""Class to control the signal repeater."""
from ..const import ROOT_SIGNAL_REPEATER
from .base_controller import BaseController
from .signal_repeater import SignalRepeater


class SignalRepeaterControl(BaseController):
    """Class to control the signal repeaters."""

    @property
    def raw(self):
        """Return raw data that it represents."""
        return self._device.raw[ROOT_SIGNAL_REPEATER]

    @property
    def signal_repeaters(self):
        """Return signal repeater objects of the signal repeater control."""
        return [SignalRepeater(self._device, i) for i in range(len(self.raw))]



"""Represent the gateway."""
from datetime import datetime

from .command import Command
from .const import (
    ATTR_AUTH,
    ATTR_COMMISSIONING_MODE,
    ATTR_CURRENT_TIME_ISO8601,
    ATTR_CURRENT_TIME_UNIX,
    ATTR_FIRMWARE_VERSION,
    ATTR_FIRST_SETUP,
    ATTR_GATEWAY_FACTORY_DEFAULTS,
    ATTR_GATEWAY_ID,
    ATTR_GATEWAY_INFO,
    ATTR_GATEWAY_REBOOT,
    ATTR_HOMEKIT_ID,
    ATTR_IDENTITY,
    ATTR_NTP,
    ATTR_PSK,
    ROOT_DEVICES,
    ROOT_GATEWAY,
    ROOT_GROUPS,
    ROOT_MOODS,
    ROOT_SMART_TASKS,
)
from .device import Device
from .group import Group
from .mood import Mood
from .smart_task import SmartTask


class Gateway:
    """This class connects to the IKEA Tradfri Gateway."""

    @classmethod
    def generate_psk(cls, identity):
        """Generate the PRE_SHARED_KEY from the gateway.

        Returns a Command.
        """

        def process_result(result):
            return result[ATTR_PSK]

        return Command(
            "post",
            [ROOT_GATEWAY, ATTR_AUTH],
            {ATTR_IDENTITY: identity},
            process_result=process_result,
        )

    @classmethod
    def get_endpoints(cls):
        """
        Return all available endpoints on the gateway.

        Returns a Command.
        """

        def process_result(result):
            return [line.split(";")[0][2:-1] for line in result.split(",")]

        return Command(
            "get",
            [".well-known", "core"],
            parse_json=False,
            process_result=process_result,
        )

    def get_devices(self):
        """
        Return the devices linked to the gateway.

        Returns a Command.
        """

        def process_result(result):
            return [self.get_device(dev) for dev in result]

        return Command("get", [ROOT_DEVICES], process_result=process_result)

    @classmethod
    def get_device(cls, device_id):
        """
        Return specified device.

        Returns a Command.
        """

        def process_result(result):
            return Device(result)

        return Command("get", [ROOT_DEVICES, device_id], process_result=process_result)

    def get_groups(self):
        """
        Return the groups linked to the gateway.

        Returns a Command.
        """

        def process_result(result):
            return [self.get_group(group) for group in result]

        return Command("get", [ROOT_GROUPS], process_result=process_result)

    def get_group(self, group_id):
        """
        Return specified group.

        Returns a Command.
        """

        def process_result(result):
            return Group(self, result)

        return Command("get", [ROOT_GROUPS, group_id], process_result=process_result)

    @classmethod
    def add_group_member(cls, values):
        """Add a device to a group."""

        return Command("put", [ROOT_GROUPS, "add"], values)

    @classmethod
    def remove_group_member(cls, values):
        """Remove a device from a group."""

        return Command("put", [ROOT_GROUPS, "remove"], values)

    @classmethod
    def get_gateway_info(cls):
        """
        Return the gateway info.

        Returns a Command.
        """

        def process_result(result):
            return GatewayInfo(result)

        return Command(
            "get", [ROOT_GATEWAY, ATTR_GATEWAY_INFO], process_result=process_result
        )

    def get_moods(self, group_id):
        """
        Return moods available in given group.

        Returns a Command.
        """

        def process_result(result):
            return [self.get_mood(mood, mood_parent=group_id) for mood in result]

        return Command("get", [ROOT_MOODS, group_id], process_result=process_result)

    @classmethod
    def get_mood(cls, mood_id, *, mood_parent=None):
        """
        Return a mood.

        Returns a Command.
        """

        def process_result(result):
            return Mood(result, mood_parent)

        return Command(
            "get",
            [ROOT_MOODS, mood_parent, mood_id],
            mood_parent,
            process_result=process_result,
        )

    def get_smart_tasks(self):
        """
        Return the transitions linked to the gateway.

        Returns a Command.
        """

        def process_result(result):
            return [self.get_smart_task(task) for task in result]

        return Command("get", [ROOT_SMART_TASKS], process_result=process_result)

    def get_smart_task(self, task_id):
        """
        Return specified transition.

        Returns a Command.
        """

        def process_result(result):
            return SmartTask(self, result)

        return Command(
            "get", [ROOT_SMART_TASKS, task_id], process_result=process_result
        )

    @classmethod
    def reboot(cls):
        """Reboot the Gateway.

        Returns a Command.
        """

        return Command("post", [ROOT_GATEWAY, ATTR_GATEWAY_REBOOT])

    @classmethod
    def set_commissioning_timeout(cls, timeout):
        """Put the gateway in pairing state.

        The pairing state is when the gateway accepts pairings from
        switches, dimmers and motion sensors for up to timeout seconds.
        Returns a Command.
        """

        return Command(
            "put", [ROOT_GATEWAY, ATTR_GATEWAY_INFO], {ATTR_COMMISSIONING_MODE: timeout}
        )

    @classmethod
    def factory_reset(cls):
        """Reset Gateway to factory defaults.

        WARNING: All data in Gateway is lost (pairing, groups, etc)
        Returns a Command.
        """

        return Command("post", [ROOT_GATEWAY, ATTR_GATEWAY_FACTORY_DEFAULTS])


class GatewayInfo:
    """This class contains Gateway information."""

    def __init__(self, raw):
        """Create object of class."""
        self.raw = raw

    @property
    def id(self):
        """Return the gateway id."""
        return self.raw.get(ATTR_GATEWAY_ID)

    @property
    def ntp_server(self):
        """NTP server in use."""
        return self.raw.get(ATTR_NTP)

    @property
    def firmware_version(self):
        """NTP server in use."""
        return self.raw.get(ATTR_FIRMWARE_VERSION)

    @property
    def current_time(self):
        """Return current time (normal timestamp)."""
        if ATTR_CURRENT_TIME_UNIX not in self.raw:
            return None
        return datetime.utcfromtimestamp(self.raw[ATTR_CURRENT_TIME_UNIX])

    @property
    def current_time_iso8601(self):
        """Return current time in iso8601 format."""
        return self.raw.get(ATTR_CURRENT_TIME_ISO8601)

    @property
    def first_setup(self):
        """Return the time when gateway was first set up."""
        if ATTR_FIRST_SETUP not in self.raw:
            return None
        return datetime.utcfromtimestamp(self.raw[ATTR_FIRST_SETUP])

    @property
    def homekit_id(self):
        """Return homekit id."""
        return self.raw.get(ATTR_HOMEKIT_ID)

    @property
    def path(self):
        """Return path."""
        return [ROOT_GATEWAY, ATTR_GATEWAY_INFO]

    def set_values(self, values):
        """Help set values for mood.

        Returns a Command.
        """
        return Command("put", self.path, values)

    def update(self):
        """
        Update the info.

        Returns a Command.
        """

        def process_result(result):
            """Define callback to process result."""
            self.raw = result

        return Command("get", self.path, process_result=process_result)

    def __repr__(self):
        """Return representation of class object."""
        return "<GatewayInfo>"



"""Group handling."""
from .color import COLORS
from .const import (
    ATTR_DEVICE_STATE,
    ATTR_GROUP_ID,
    ATTR_GROUP_MEMBERS,
    ATTR_HS_LINK,
    ATTR_ID,
    ATTR_LIGHT_COLOR_HEX,
    ATTR_LIGHT_COLOR_HUE,
    ATTR_LIGHT_COLOR_SATURATION,
    ATTR_LIGHT_COLOR_X,
    ATTR_LIGHT_COLOR_Y,
    ATTR_LIGHT_DIMMER,
    ATTR_LIGHT_MIREDS,
    ATTR_MOOD,
    ATTR_TRANSITION_TIME,
    RANGE_BRIGHTNESS,
    RANGE_HUE,
    RANGE_MIREDS,
    RANGE_SATURATION,
    RANGE_X,
    RANGE_Y,
    ROOT_GROUPS,
)
from .error import ColorError
from .resource import ApiResource


class Group(ApiResource):
    """Represent a group."""

    def __init__(self, gateway, raw):
        """Create object of class."""
        super().__init__(raw)
        self._gateway = gateway

    @property
    def path(self):
        """Path."""
        return [ROOT_GROUPS, self.id]

    @property
    def state(self):
        """Boolean representing the light state of the group."""
        return self.raw.get(ATTR_DEVICE_STATE) == 1

    @property
    def dimmer(self):
        """Dimmer value of the group."""
        return self.raw.get(ATTR_LIGHT_DIMMER)

    @property
    def hex_color(self):
        """Return hex color."""
        return self.raw.get(ATTR_LIGHT_COLOR_HEX)

    @property
    def member_ids(self):
        """Members of this group."""
        info = self.raw.get(ATTR_GROUP_MEMBERS, {})

        if not info or ATTR_HS_LINK not in info:
            return []

        return info[ATTR_HS_LINK].get(ATTR_ID, [])

    @property
    def mood_id(self):
        """Active mood."""
        return self.raw.get(ATTR_MOOD)

    def members(self):
        """Return device objects of members of this group."""
        return [self._gateway.get_device(dev) for dev in self.member_ids]

    def add_member(self, memberid):
        """Add a member to this group."""
        return self._gateway.add_group_member(
            {ATTR_GROUP_ID: self.id, ATTR_ID: [memberid]}
        )

    def remove_member(self, memberid):
        """Remove a member from this group."""
        return self._gateway.remove_group_member(
            {ATTR_GROUP_ID: self.id, ATTR_ID: [memberid]}
        )

    def moods(self):
        """Return mood objects of moods in this group."""
        return self._gateway.get_moods(self.id)

    def mood(self):
        """Active mood."""
        return self._gateway.get_mood(self.mood_id, mood_parent=self.id)

    def activate_mood(self, mood_id):
        """Activate a mood."""
        return self.set_values({ATTR_MOOD: mood_id, ATTR_DEVICE_STATE: int(self.state)})

    def set_state(self, state):
        """Set state of a group."""
        return self.set_values({ATTR_DEVICE_STATE: int(state)})

    def set_dimmer(self, dimmer, transition_time=None):
        """Set dimmer value of a group.

        dimmer: Integer between 0..255
        transition_time: Integer representing tenth of a second (default None)
        """
        values = {
            ATTR_LIGHT_DIMMER: dimmer,
        }
        if transition_time is not None:
            values[ATTR_TRANSITION_TIME] = transition_time
        return self.set_values(values)

    def set_color_temp(self, color_temp, *, index=0, transition_time=None):
        """Set color temp a light."""
        self._value_validate(color_temp, RANGE_MIREDS, "Color temperature")

        values = {ATTR_LIGHT_MIREDS: color_temp}

        if transition_time is not None:
            values[ATTR_TRANSITION_TIME] = transition_time

        return self.set_values(values)

    def set_hex_color(self, color, transition_time=None):
        """Set hex color of a group."""
        values = {
            ATTR_LIGHT_COLOR_HEX: color,
        }
        if transition_time is not None:
            values[ATTR_TRANSITION_TIME] = transition_time
        return self.set_values(values)

    def set_hsb(
        self, hue, saturation, brightness=None, *, index=0, transition_time=None
    ):
        """Set HSB color settings of the light."""
        self._value_validate(hue, RANGE_HUE, "Hue")
        self._value_validate(saturation, RANGE_SATURATION, "Saturation")

        values = {ATTR_LIGHT_COLOR_SATURATION: saturation, ATTR_LIGHT_COLOR_HUE: hue}

        if brightness is not None:
            values[ATTR_LIGHT_DIMMER] = brightness
            self._value_validate(brightness, RANGE_BRIGHTNESS, "Brightness")

        if transition_time is not None:
            values[ATTR_TRANSITION_TIME] = transition_time

        return self.set_values(values)

    def set_xy_color(self, color_x, color_y, transition_time=None):
        """Set xy color of a group."""
        self._value_validate(color_x, RANGE_X, "X color")
        self._value_validate(color_y, RANGE_Y, "Y color")

        values = {ATTR_LIGHT_COLOR_X: color_x, ATTR_LIGHT_COLOR_Y: color_y}

        if transition_time is not None:
            values[ATTR_TRANSITION_TIME] = transition_time

        return self.set_values(values)

    def set_predefined_color(self, colorname, transition_time=None):
        """Set predefined color for group."""
        try:
            color = COLORS[colorname.lower().replace(" ", "_")]
            return self.set_hex_color(color, transition_time=transition_time)
        except KeyError as exc:
            raise ColorError(f"Invalid color specified: {colorname}") from exc

    def _value_validate(
        self, value, rnge, identifier="Given"
    ):  # pylint: disable=no-self-use
        """Make sure a value is within a given range."""
        if value is not None and (value < rnge[0] or value > rnge[1]):
            raise ValueError(
                f"{identifier} value must be between {rnge[0]} and {rnge[1]}."
            )

    def __repr__(self):
        """Return representation of class object."""
        state = "on" if self.state else "off"
        return f"<Group {self.name} - {state}>"



"""Errors for PyTradfri."""


class PytradfriError(Exception):
    """Base Error."""


class RequestError(PytradfriError):
    """An error happened sending or receiving a command."""


class ColorError(PytradfriError):
    """An error happened matching color name."""


class RequestTimeout(RequestError):
    """Error when sending or receiving the command timed out."""


class ClientError(RequestError):
    """Error when the client caused the error.

    See section 5.9.2 of draft-ietf-core-coap-04.
    """


class ServerError(RequestError):
    """Error when the server caused the error.

    See section 5.9.3 of draft-ietf-core-coap-04.
    """



"""Represent a blind."""
from __future__ import annotations

from typing import TYPE_CHECKING, cast

from ..const import ATTR_BLIND_CURRENT_POSITION, ATTR_START_BLINDS
from ..resource import TypeRaw, TypeRawList

if TYPE_CHECKING:
    # avoid cyclic import at runtime.
    from . import Device


class Blind:
    """Represent a blind."""

    def __init__(self, device: Device, index: int) -> None:
        """Create object of class."""
        self.device = device
        self.index = index

    @property
    def raw(self) -> TypeRaw:
        """Return raw data that it represents."""
        return cast(
            TypeRaw,
            cast(TypeRawList, self.device.raw)[ATTR_START_BLINDS][self.index],
        )

    @property
    def current_cover_position(self) -> int:
        """Get the current position of the blind."""
        return cast(int, self.raw.get(ATTR_BLIND_CURRENT_POSITION))



"""Constants."""
from typing import Final

ROOT_DEVICES = "15001"
ROOT_GATEWAY = "15011"
ROOT_GROUPS = "15004"
ROOT_MOODS = "15005"
ROOT_NOTIFICATION = "15006"  # speculative name
ROOT_REMOTE_CONTROL = "15009"
ROOT_SIGNAL_REPEATER = "15014"
ROOT_SMART_TASKS = "15010"
ROOT_START_ACTION = "15013"  # found under ATTR_START_ACTION
ATTR_START_BLINDS = "15015"
ROOT_AIR_PURIFIER = "15025"

ATTR_ALEXA_PAIR_STATUS = "9093"
ATTR_AUTH = "9063"
ATTR_APPLICATION_TYPE = "5750"
ATTR_APPLICATION_TYPE_BLIND = 7

ATTR_BLIND_CURRENT_POSITION = "5536"
ATTR_BLIND_TRIGGER = "5523"

ATTR_AIR_PURIFIER_MODE: Final = "5900"
ATTR_AIR_PURIFIER_FILTER_RUNTIME: Final = "5902"
ATTR_AIR_PURIFIER_FILTER_STATUS: Final = "5903"
ATTR_AIR_PURIFIER_FILTER_LIFETIME_TOTAL: Final = "5904"
ATTR_AIR_PURIFIER_CONTROLS_LOCKED: Final = "5905"
ATTR_AIR_PURIFIER_LEDS_OFF: Final = "5906"
ATTR_AIR_PURIFIER_AIR_QUALITY: Final = "5907"
ATTR_AIR_PURIFIER_FAN_SPEED: Final = "5908"
ATTR_AIR_PURIFIER_MOTOR_RUNTIME_TOTAL: Final = "5909"
ATTR_AIR_PURIFIER_FILTER_LIFETIME_REMAINING: Final = "5910"
ATTR_AIR_PURIFIER_MODE_AUTO: Final = 1

ATTR_CERTIFICATE_PEM = "9096"
ATTR_CERTIFICATE_PROV = "9092"
ATTR_CLIENT_IDENTITY_PROPOSED = "9090"
ATTR_CREATED_AT = "9002"
ATTR_COGNITO_ID = "9101"
ATTR_COMMISSIONING_MODE = "9061"
ATTR_CURRENT_TIME_UNIX = "9059"
ATTR_CURRENT_TIME_ISO8601 = "9060"

ATTR_DEVICE_INFO = "3"

ATTR_GATEWAY_ID_2 = "9100"  # stored in IKEA app code as gateway id
ATTR_GATEWAY_TIME_SOURCE = "9071"
ATTR_GATEWAY_UPDATE_PROGRESS = "9055"

ATTR_GROUP_MEMBERS = "9018"
ATTR_GROUP_ID = "9038"

ATTR_HOMEKIT_ID = "9083"
ATTR_HS_LINK = "15002"

ATTR_ID = "9003"
ATTR_IDENTITY = "9090"
ATTR_IOT_ENDPOINT = "9103"

ATTR_KEY_PAIR = "9097"

ATTR_LAST_SEEN = "9020"
ATTR_LIGHT_CONTROL = "3311"  # array

ATTR_MASTER_TOKEN_TAG = "9036"
ATTR_MOOD = "9039"

ATTR_NAME = "9001"
ATTR_NTP = "9023"
ATTR_FIRMWARE_VERSION = "9029"
ATTR_FIRST_SETUP = "9069"  # ??? unix epoch value when gateway first setup

ATTR_GATEWAY_INFO = "15012"
ATTR_GATEWAY_ID = "9081"  # ??? id of the gateway
ATTR_GATEWAY_REBOOT = "9030"  # gw reboot
ATTR_GATEWAY_FACTORY_DEFAULTS = "9031"  # gw to factory defaults
ATTR_GATEWAY_FACTORY_DEFAULTS_MIN_MAX_MSR = "5605"
ATTR_GOOGLE_HOME_PAIR_STATUS = "9105"

ATTR_DEVICE_STATE = "5850"  # 0 / 1
ATTR_LIGHT_DIMMER = "5851"  # Dimmer, not following spec: 0..255
ATTR_LIGHT_COLOR_HEX = "5706"  # string representing a value in hex
ATTR_LIGHT_COLOR_X = "5709"
ATTR_LIGHT_COLOR_Y = "5710"
ATTR_LIGHT_COLOR_HUE = "5707"
ATTR_LIGHT_COLOR_SATURATION = "5708"
ATTR_LIGHT_MIREDS = "5711"

ATTR_NOTIFICATION_EVENT = "9015"
ATTR_NOTIFICATION_NVPAIR = "9017"
ATTR_NOTIFICATION_STATE = "9014"

ATTR_OTA_TYPE = "9066"
ATTR_OTA_UPDATE_STATE = "9054"
ATTR_OTA_UPDATE = "9037"

ATTR_PUBLIC_KEY = "9098"
ATTR_PRIVATE_KEY = "9099"
ATTR_PSK = "9091"

ATTR_REACHABLE_STATE = "9019"
ATTR_REPEAT_DAYS = "9041"

ATTR_SEND_CERT_TO_GATEWAY = "9094"
ATTR_SEND_COGNITO_ID_TO_GATEWAY = "9095"
ATTR_SEND_GH_COGNITO_ID_TO_GATEWAY = "9104"
ATTR_SENSOR = "3300"
ATTR_SENSOR_MAX_RANGE_VALUE = "5604"
ATTR_SENSOR_MAX_MEASURED_VALUE = "5602"
ATTR_SENSOR_MIN_RANGE_VALUE = "5603"
ATTR_SENSOR_MIN_MEASURED_VALUE = "5601"
ATTR_SENSOR_TYPE = "5751"
ATTR_SENSOR_UNIT = "5701"
ATTR_SENSOR_VALUE = "5700"
ATTR_START_ACTION = "9042"  # array
ATTR_SMART_TASK_TYPE = "9040"  # 4 = transition | 1 = not home | 2 = on/off
ATTR_SMART_TASK_NOT_AT_HOME = 1
ATTR_SMART_TASK_LIGHTS_OFF = 2
ATTR_SMART_TASK_WAKE_UP = 4
ATTR_SMART_TASK_TRIGGER_TIME_INTERVAL = "9044"
ATTR_SMART_TASK_TRIGGER_TIME_START_HOUR = "9046"
ATTR_SMART_TASK_TRIGGER_TIME_START_MIN = "9047"

ATTR_SWITCH_CUM_ACTIVE_POWER = "5805"
ATTR_SWITCH_ON_TIME = "5852"
ATTR_SWITCH_PLUG = "3312"
ATTR_SWITCH_POWER_FACTOR = "5820"

ATTR_TIME_END_TIME_HOUR = "9048"
ATTR_TIME_END_TIME_MINUTE = "9049"
ATTR_TIME_START_TIME_HOUR = "9046"
ATTR_TIME_START_TIME_MINUTE = "9047"

ATTR_TRANSITION_TIME = "5712"

ATTR_USE_CURRENT_LIGHT_SETTINGS = "9070"

# URL to json-file containing links to all firmware updates
URL_OTA_FW = "http://fw.ota.homesmart.ikea.net/feed/version_info.json"

# Mireds range that white-spectrum bulbs can show
RANGE_MIREDS = (250, 454)

# Hue of a RGB bulb
RANGE_HUE = (0, 65535)
# Effective saturation range of a RGB bulb. The bulb will accept
# slightly higher values, but it won't produce any light.
RANGE_SATURATION = (0, 65279)
# Brightness range of all bulbs. 0 will turn off the lamp
RANGE_BRIGHTNESS = (0, 254)

RANGE_BLIND = (0, 100)

RANGE_AIR_PURIFIER = (2, 50)

# XY color
RANGE_X = (0, 65535)
RANGE_Y = (0, 65535)

SUPPORT_BRIGHTNESS = 1
SUPPORT_COLOR_TEMP = 2
SUPPORT_HEX_COLOR = 4
SUPPORT_RGB_COLOR = 8
SUPPORT_XY_COLOR = 16

ATTR_DEVICE_MANUFACTURER: Final = "0"
ATTR_DEVICE_MODEL_NUMBER: Final = "1"
ATTR_DEVICE_SERIAL: Final = "2"
ATTR_DEVICE_FIRMWARE_VERSION: Final = "3"
ATTR_DEVICE_POWER_SOURCE: Final = "6"
ATTR_DEVICE_BATTERY: Final = "9"



"""Class to control the air purifiers."""
from __future__ import annotations

from typing import cast

from ..command import Command
from ..const import (
    ATTR_AIR_PURIFIER_CONTROLS_LOCKED,
    ATTR_AIR_PURIFIER_LEDS_OFF,
    ATTR_AIR_PURIFIER_MODE,
    ATTR_AIR_PURIFIER_MODE_AUTO,
    RANGE_AIR_PURIFIER,
    ROOT_AIR_PURIFIER,
)
from ..resource import TypeRaw
from .air_purifier import AirPurifier
from .base_controller import BaseController


class AirPurifierControl(BaseController):
    """Class to control the air purifiers."""

    @property
    def raw(self) -> TypeRaw:
        """Return raw data that it represents."""
        return cast(TypeRaw, self._device.raw[ROOT_AIR_PURIFIER])

    @property
    def air_purifiers(self) -> list[AirPurifier]:
        """Return air purifier objects of the air purifier control."""
        return [AirPurifier(self._device, i) for i in range(len(self.raw))]

    def turn_off(self, *, index=0) -> Command:
        """Turn the device off."""
        return self.set_value({ATTR_AIR_PURIFIER_MODE: 0}, index=index)

    def turn_on_auto_mode(self, *, index=0) -> Command:
        """Turn on auto mode."""
        return self.set_value(
            {ATTR_AIR_PURIFIER_MODE: ATTR_AIR_PURIFIER_MODE_AUTO}, index=index
        )

    def set_fan_speed(self, mode: int, *, index=0) -> Command:
        """Set the fan speed of the purifier."""
        self._value_validate(mode, RANGE_AIR_PURIFIER, "Air Purifier mode")
        return self.set_value({ATTR_AIR_PURIFIER_MODE: mode}, index=index)

    def set_controls_locked(self, locked: bool, *, index=0) -> Command:
        """Set physical controls locked of the air purifier."""

        return self.set_value(
            {ATTR_AIR_PURIFIER_CONTROLS_LOCKED: 1 if locked else 0}, index=index
        )

    def set_leds_off(self, leds_off: bool, *, index=0) -> Command:
        """Set led's off/on of the air purifier."""

        return self.set_value(
            {ATTR_AIR_PURIFIER_LEDS_OFF: 1 if leds_off else 0}, index=index
        )

    def set_value(self, value: dict[str, bool | int], *, index=0) -> Command:
        """Set values on air purifier control.

        Returns a Command.
        """
        assert len(self.raw) == 1, "Only devices with 1 air purifier supported"

        return Command("put", self._device.path, {ROOT_AIR_PURIFIER: [value]})



"""JSON utility functions."""
from __future__ import annotations

import json
import logging

from .error import PytradfriError

#  https://github.com/home-assistant/home-assistant/blob/4e8723f345d526ffbcbea74444e1a140a7eec863/homeassistant/util/json.py


_LOGGER = logging.getLogger(__name__)


def load_json(filename: str) -> list | dict:
    """Load JSON data from a file and return as dict or list.

    Defaults to returning empty dict if file is not found.
    """
    try:
        with open(filename, encoding="utf-8") as fdesc:
            return json.loads(fdesc.read())
    except FileNotFoundError:
        # This is not a fatal error
        _LOGGER.debug("JSON file not found: %s", filename)
    except ValueError as exc:
        _LOGGER.exception("Could not parse JSON content: %s", filename)
        raise PytradfriError(exc) from exc
    except OSError as exc:
        _LOGGER.exception("JSON file reading failed: %s", filename)
        raise PytradfriError(exc) from exc
    return {}  # (also evaluates to False)


def save_json(filename: str, config: list | dict) -> bool:
    """Save JSON data to a file.

    Returns True on success.
    """
    try:
        data = json.dumps(config, sort_keys=True, indent=4)
        with open(filename, "w", encoding="utf-8") as fdesc:
            fdesc.write(data)
            return True
    except TypeError as exc:
        _LOGGER.exception("Failed to serialize to JSON: %s", filename)
        raise PytradfriError(exc) from exc
    except OSError as exc:
        _LOGGER.exception("Saving JSON file failed: %s", filename)
        raise PytradfriError(exc) from exc


class BitChoices:
    """Helper class for bitwise dates.

    http://stackoverflow.com/questions/3663898/representing-a-multi-select-field-for-weekdays-in-a-django-model
    """

    def __init__(self, choices):
        """Initialize BitChoices class."""
        self._choices = []
        self._lookup = {}
        for index, (key, val) in enumerate(choices):
            index = 2**index
            self._choices.append((index, val))
            self._lookup[key] = index

    def __iter__(self):
        """Iterate over object."""
        return iter(self._choices)

    def __len__(self):
        """Len."""
        return len(self._choices)

    def __getattr__(self, attr):
        """Getattr."""
        try:
            return self._lookup[attr]
        except KeyError as exc:
            raise AttributeError(attr) from exc

    def get_selected_keys(self, selection):
        """Return a list of keys for the given selection."""
        return [k for k, b in self._lookup.items() if b & selection]

    def get_selected_values(self, selection):
        """Return a list of values for the given selection."""
        return [v for b, v in self._choices if b & selection]



"""Classes to interact with devices."""
from datetime import datetime
from typing import Optional, TypedDict, cast

from ..const import (
    ATTR_APPLICATION_TYPE,
    ATTR_DEVICE_BATTERY,
    ATTR_DEVICE_FIRMWARE_VERSION,
    ATTR_DEVICE_INFO,
    ATTR_DEVICE_MANUFACTURER,
    ATTR_DEVICE_MODEL_NUMBER,
    ATTR_DEVICE_POWER_SOURCE,
    ATTR_DEVICE_SERIAL,
    ATTR_LAST_SEEN,
    ATTR_LIGHT_CONTROL,
    ATTR_REACHABLE_STATE,
    ATTR_START_BLINDS,
    ATTR_SWITCH_PLUG,
    ROOT_AIR_PURIFIER,
    ROOT_DEVICES,
    ROOT_SIGNAL_REPEATER,
)
from ..resource import ApiResource
from .air_purifier_control import AirPurifierControl
from .blind_control import BlindControl
from .light_control import LightControl
from .signal_repeater_control import SignalRepeaterControl
from .socket_control import SocketControl

TypeDeviceInfo = TypedDict(
    # Alternative syntax required due to the need of using strings as keys:
    # https://www.python.org/dev/peps/pep-0589/#alternative-syntax
    "TypeDeviceInfo",
    {
        "0": str,  # Manufacturer
        "1": str,  # Model number
        "2": str,  # Serial number
        "3": str,  # Firmware version
        "6": int,  # Power source
        "7": str,  # OTA image type
        "9": int,  # Battery level
    },
)


class Device(ApiResource):
    """Base class for devices."""

    @property
    def application_type(self):
        """Return application type."""
        return self.raw.get(ATTR_APPLICATION_TYPE)

    @property
    def path(self):
        """Return path."""
        return [ROOT_DEVICES, self.id]

    @property
    def device_info(self):
        """Return Device information."""
        return DeviceInfo(self)

    @property
    def last_seen(self):
        """Return timestamp when last seen."""
        if ATTR_LAST_SEEN not in self.raw:
            return None
        return datetime.utcfromtimestamp(self.raw[ATTR_LAST_SEEN])

    @property
    def reachable(self):
        """Check if gateway is reachable."""
        return self.raw.get(ATTR_REACHABLE_STATE) == 1

    @property
    def has_light_control(self):
        """Check if light_control is present."""
        return self.raw is not None and len(self.raw.get(ATTR_LIGHT_CONTROL, "")) > 0

    @property
    def light_control(self):
        """Return light_control."""
        return LightControl(self)

    @property
    def has_socket_control(self):
        """Check if socket_control is present."""
        return self.raw is not None and len(self.raw.get(ATTR_SWITCH_PLUG, "")) > 0

    @property
    def socket_control(self):
        """Return socket_control."""
        if self.has_socket_control:
            return SocketControl(self)
        return None

    @property
    def has_blind_control(self):
        """Check if blind_control is present."""
        return self.raw is not None and len(self.raw.get(ATTR_START_BLINDS, "")) > 0

    @property
    def blind_control(self):
        """Return blind_control."""
        if self.has_blind_control:
            return BlindControl(self)
        return None

    @property
    def has_signal_repeater_control(self):
        """Check if signal_repeater_control is present."""
        return self.raw is not None and len(self.raw.get(ROOT_SIGNAL_REPEATER, "")) > 0

    @property
    def signal_repeater_control(self):
        """Return signal_repeater control, if any."""
        if self.has_signal_repeater_control:
            return SignalRepeaterControl(self)
        return None

    @property
    def has_air_purifier_control(self):
        """Check if air_purifier_control is present."""
        return self.raw is not None and len(self.raw.get(ROOT_AIR_PURIFIER, "")) > 0

    @property
    def air_purifier_control(self):
        """Return air_purifier control, if any."""
        if self.has_air_purifier_control:
            return AirPurifierControl(self)
        return None

    def __repr__(self):
        """Return representation of class object."""
        return f"<{self.id} - {self.name} ({self.device_info.model_number})>"


class DeviceInfo:
    """Represent device information.

    http://www.openmobilealliance.org/tech/profiles/LWM2M_Device-v1_0.xml
    """

    VALUE_POWER_SOURCES = {
        1: "Internal Battery",
        2: "External Battery",
        3: "Battery",  # Not in spec, used by remote
        4: "Power over Ethernet",
        5: "USB",
        6: "AC (Mains) power",
        7: "Solar",
    }

    def __init__(self, device: Device) -> None:
        """Create object of class."""
        self._device = device

    @property
    def manufacturer(self) -> str:
        """Human readable manufacturer name."""
        return self.raw[ATTR_DEVICE_MANUFACTURER]

    @property
    def model_number(self) -> str:
        """Return model identifier string (manufacturer specified string)."""
        return self.raw[ATTR_DEVICE_MODEL_NUMBER]

    @property
    def serial(self) -> str:
        """Return serial string."""
        return self.raw[ATTR_DEVICE_SERIAL]

    @property
    def firmware_version(self) -> str:
        """Return current firmware version of device."""
        return self.raw[ATTR_DEVICE_FIRMWARE_VERSION]

    @property
    def power_source(self) -> int:
        """Power source."""
        return self.raw[ATTR_DEVICE_POWER_SOURCE]

    @property
    def power_source_str(self) -> Optional[str]:
        """Represent current power source."""
        if "6" not in self.raw:
            return None

        return DeviceInfo.VALUE_POWER_SOURCES.get(self.power_source, "Unknown")

    @property
    def battery_level(self) -> Optional[int]:
        """Battery in 0..100."""
        if ATTR_DEVICE_BATTERY not in self.raw:
            return None

        return self.raw[ATTR_DEVICE_BATTERY]

    @property
    def raw(self) -> TypeDeviceInfo:
        """Return raw data that it represents."""
        return cast(TypeDeviceInfo, self._device.raw[ATTR_DEVICE_INFO])



"""Base class for a controller."""
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    # avoid cyclic import at runtime.
    from . import Device


class BaseController:
    """Represent a controller."""

    def __init__(self, device: Device) -> None:
        """Create object of class."""
        self._device = device

    @property
    def raw(self):
        """Return raw data that it represents."""
        return None

    @classmethod
    def _value_validate(
        cls, value: int, rnge: list[int] | tuple[int, int], identifier: str = "Given"
    ) -> None:
        """Make sure a value is within a given range."""
        if value is not None and (value < rnge[0] or value > rnge[1]):
            raise ValueError(
                f"{identifier} value must be between {rnge[0]} and {rnge[1]}."
            )

    def __repr__(self):
        """Return representation of class object."""
        return (
            f"<{type(self).__name__} for {self._device.name} ({len(self.raw)} devices)>"
        )



"""Provide a CLI for Tradfri."""
import argparse
import logging
from pprint import pprint
import uuid

from pytradfri.api.libcoap_api import APIFactory
from pytradfri.error import PytradfriError
from pytradfri.util import load_json, save_json

from .command import Command
from .gateway import Gateway

CONFIG_FILE = "tradfri_standalone_psk.conf"


def main(org_args):
    """Run main."""
    global devices, homekit_id, light, api, lights, gateway, groups, moods, tasks, dump_devices, dump_all  # pylint: disable=global-variable-undefined, invalid-name

    if org_args.host not in load_json(CONFIG_FILE) and org_args.key is None:
        print(
            "Please provide the 'Security Code' on the back of your "
            "Tradfri gateway:",
            end=" ",
        )
        key = input().strip()
        if len(key) != 16:
            raise PytradfriError(
                "'Security Code' has to be exactly" + "16 characters long."
            )
        org_args.key = key

    conf = load_json(CONFIG_FILE)

    try:
        identity = conf[org_args.host].get("identity")
        psk = conf[org_args.host].get("key")
        api_factory = APIFactory(host=org_args.host, psk_id=identity, psk=psk)
    except KeyError:
        identity = uuid.uuid4().hex
        api_factory = APIFactory(host=org_args.host, psk_id=identity)

        try:
            psk = api_factory.generate_psk(org_args.key)
            print("Generated PSK: ", psk)

            conf[org_args.host] = {"identity": identity, "key": psk}
            save_json(CONFIG_FILE, conf)
        except AttributeError as exc:
            raise PytradfriError(
                "Please provide the 'Security Code' on the "
                "back of your Tradfri gateway using the "
                "-K flag."
            ) from exc

    api = api_factory.request

    gateway = Gateway()
    devices_commands = api(gateway.get_devices())
    devices = api(devices_commands)
    lights = [dev for dev in devices if dev.has_light_control]
    if lights:
        light = lights[0]
    else:
        print("No lights found!")
        light = None
    groups_commands = api(gateway.get_groups())
    groups = api(groups_commands)
    moods = []
    if groups:
        group = groups[0]
        for group in groups:
            moods_commands = api(group.moods())
            group_moods = api(moods_commands)
            moods.extend(group_moods)
    else:
        print("No groups found!")
        group = None
    tasks_commands = api(gateway.get_smart_tasks())
    tasks = api(tasks_commands)
    homekit_id = api(gateway.get_gateway_info()).homekit_id

    def dump_all():  # pylint: disable=unused-variable
        """Dump all endpoints."""
        endpoints = api(gateway.get_endpoints())

        for endpoint in endpoints:
            parts = endpoint[1:].split("/")

            if not all(part.isdigit() for part in parts):
                continue

            pprint(api(Command("get", parts)))
            print()
            print()

    def dump_devices():  # pylint: disable=unused-variable
        """Dump devices."""
        pprint([d.raw for d in devices])

    print()
    print("Example commands:")
    print("> devices")
    print("> homekit_id")
    if light:
        print("> light.light_control.lights")
        print("> api(light.light_control.set_dimmer(10))")
        print("> api(light.light_control.set_dimmer(254))")
        print("> api(light.light_control.set_xy_color(30015, 26870))")
        print("> api(light.light_control.set_predefined_color('Warm Amber'))")
        print("> api(lights[1].light_control.set_dimmer(20))")
    if tasks:
        print("> tasks[0].repeat_days_list")
    print("> api(gateway.reboot())")
    print("> groups")
    print("> moods")
    print("> tasks")
    print("> dump_devices()")
    print("> dump_all()")


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "host", metavar="IP", type=str, help="IP Address of your Tradfri gateway"
    )
    parser.add_argument(
        "-K",
        "--key",
        dest="key",
        required=False,
        help="Security code found on your Tradfri gateway",
    )
    args = parser.parse_args()
    main(args)



"""Represent a signal repeater."""
from ..const import ROOT_SIGNAL_REPEATER


class SignalRepeater:
    """Represent a signal repeater."""

    def __init__(self, device, index):
        """Create object of class."""
        self.device = device
        self.index = index

    @property
    def raw(self):
        """Return raw data that it represents."""
        return self.device.raw[ROOT_SIGNAL_REPEATER][self.index]
